import matplotlib

matplotlib.use("Agg")

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest

from medevalkit.plots import (
    plot_calibration_curve,
    plot_decision_curve,
    plot_multiclass_calibration,
    plot_multiclass_roc,
    plot_roc_curve,
    plot_threshold_metrics,
)


@pytest.fixture(autouse=True)
def no_show(monkeypatch):
    monkeypatch.setattr(plt, "show", lambda *args, **kwargs: None)
    yield
    plt.close("all")


class _DummyCalibration:
    def __init__(self):
        grid = np.linspace(0.1, 0.9, 5)
        self._curves = {
            0: {
                "brier_score": 0.1,
                "calibration_curve": {"prob_pred": grid, "prob_true": grid},
            },
            1: {
                "brier_score": 0.2,
                "calibration_curve": {"prob_pred": grid, "prob_true": grid ** 0.9},
            },
        }

    def one_vs_rest_curves(self):
        return self._curves


def _assert_figure_created():
    assert len(plt.get_fignums()) >= 1


def test_binary_plot_functions_execute_without_error():
    y_true = np.array([0, 1, 1, 0, 1, 0])
    y_prob = np.array([0.1, 0.8, 0.6, 0.3, 0.9, 0.2])

    plot_roc_curve(y_true, y_prob)
    _assert_figure_created()
    plt.close("all")

    plot_calibration_curve(y_true, y_prob, n_bins=3)
    _assert_figure_created()
    plt.close("all")

    thresholds = {
        0.0: {"sensitivity": 1.0, "specificity": 0.0},
        0.5: {"sensitivity": 0.8, "specificity": 0.7},
        1.0: {"sensitivity": 0.0, "specificity": 1.0},
    }
    plot_threshold_metrics(thresholds)
    _assert_figure_created()
    plt.close("all")

    dca_df = pd.DataFrame({
        "threshold": np.linspace(0.1, 0.9, 5),
        "net_benefit": np.linspace(0.05, 0.15, 5),
    })
    plot_decision_curve(dca_df)
    _assert_figure_created()
    plt.close("all")


def test_multiclass_plot_functions_execute_without_error():
    y_true = np.array([0, 1, 2, 1, 0])
    y_prob = np.array(
        [
            [0.7, 0.2, 0.1],
            [0.1, 0.7, 0.2],
            [0.1, 0.2, 0.7],
            [0.2, 0.6, 0.2],
            [0.8, 0.1, 0.1],
        ]
    )

    plot_multiclass_roc(y_true, y_prob)
    _assert_figure_created()
    plt.close("all")

    dummy_calibration = _DummyCalibration()
    plot_multiclass_calibration(dummy_calibration)
    _assert_figure_created()
    plt.close("all")



"""Calibration metrics and plotting for classification models."""

import numpy as np
from numpy.typing import ArrayLike
from sklearn.calibration import calibration_curve
from sklearn.metrics import brier_score_loss
from .bootstrap import Bootstrapper

class BaseCalibration:
    
    def __init__(self, y_true: ArrayLike, y_pred_prob: ArrayLike, n_bins: int = 10):
        self.y_true = y_true
        self.y_pred_prob = y_pred_prob
        self.n_bins = n_bins

class BinaryCalibration(BaseCalibration):

    def __init__(self, y_true: ArrayLike, y_pred_prob: ArrayLike, n_bins: int = 10):
        super().__init__(y_true = y_true, y_pred_prob = y_pred_prob, n_bins = n_bins)

    def compute(self, bootstrap: bool = True, **kwargs) -> dict:
        prob_true, prob_pred = calibration_curve(self.y_true, self.y_pred_prob, n_bins=self.n_bins)
        output = {
            'brier_score': brier_score_loss(self.y_true, self.y_pred_prob),
            'calibration_curve': (prob_pred, prob_true)
        }
        if bootstrap:
            bs = Bootstrapper(**kwargs)
            brier, brier_lower, brier_upper = bs.bootstrap_ci(brier_score_loss,self.y_true,self.y_pred_prob)
            output.update({
                'brier_score': brier,
                'brier_score_upper': brier_upper,
                'brier_score_lower': brier_lower,
                'calibration_curve': {'prob_pred':prob_pred,'prob_true':prob_true}
            })
        
        return output

class MulticlassCalibration(BaseCalibration):

    def __init__(self, y_true: ArrayLike, y_pred_prob: ArrayLike, n_bins: int = 10):
        super().__init__(y_true = y_true, y_pred_prob = y_pred_prob, n_bins = n_bins)
        self.n_classes = y_pred_prob.shape[1]

    def one_vs_rest_curves(self, bootstrap: bool = True, **kwargs) -> dict: 
        """
        Returns calibration curves and Brier scores for each class in a one-vs-rest fashion.
        Output: {
            class_index: {
                'prob_true': [...],
                'prob_pred': [...],
                'brier_score': float
            },
            ...
        }
        """
        curves = {}
        for k in range(self.n_classes):
            y_true_bin = (self.y_true == k).astype(int)
            prob_k = self.y_pred_prob[:, k]
            prob_true, prob_pred = calibration_curve(y_true_bin, prob_k, n_bins=self.n_bins)
            brier = brier_score_loss(y_true_bin, prob_k)
            curves[k] = {
                'brier_score': brier,
                'calibration_curve': {'prob_pred':prob_pred,'prob_true':prob_true}
            }
            if bootstrap:
                bs = Bootstrapper(**kwargs)
                brier, brier_lower, brier_upper = bs.bootstrap_ci(brier_score_loss, y_true_bin, prob_k)
                curves[k].update({
                'brier_score': brier,
                'brier_score_upper': brier_upper,
                'brier_score_lower': brier_lower
            })
        return curves

    @staticmethod
    def _expected_calibration_error(y_true: ArrayLike, y_pred_prob: ArrayLike, n_bins = 10) -> float:
        """
        Computes Expected Calibration Error (ECE) for multiclass classification.
        """
        y_pred = np.argmax(y_pred_prob, axis=1)
        confidences = np.max(y_pred_prob, axis=1)
        accuracies = (y_pred == y_true).astype(int)

        bins = np.linspace(0, 1, n_bins + 1)
        ece = 0.0
        for i in range(n_bins):
            bin_mask = (confidences > bins[i]) & (confidences <= bins[i+1])
            bin_size = np.sum(bin_mask)
            if bin_size > 0:
                acc = np.mean(accuracies[bin_mask])
                conf = np.mean(confidences[bin_mask])
                ece += (bin_size / len(y_true)) * abs(acc - conf)
        return ece
    
    def expected_calibration_error(self,bootstrap: bool = True, **kwargs):
        output = {
            'expected_calibration_error': self._expected_calibration_error(self.y_true,self.y_pred_prob)
        }
        if bootstrap:
            bs = Bootstrapper(**kwargs)
            ece, ece_lower, ece_upper = bs.bootstrap_ci(self._expected_calibration_error,self.y_true,self.y_pred_prob)
            output.update({
                'expected_calibration_error': ece,
                'expected_calibration_error_upper': ece_upper,
                'expected_calibration_error_lower': ece_lower
            })
            
        return output


"""Bootstrap resampling utilities for confidence interval estimation."""

import types
import numpy as np
from numpy.typing import ArrayLike

class Bootstrapper:

    def __init__(self, n_resamples: int = 1000, random_state : int = 1):
        self.n_resamples = n_resamples
        self.random_state = np.random.RandomState(random_state)
    
    def bootstrap(self, metric_fn: types.FunctionType, y_true: ArrayLike, y_pred: ArrayLike, **kwargs) -> list:
        y_true = np.asarray(y_true)
        y_pred = np.asarray(y_pred)
        bootstrapped_scores = []
        n_samples = len(y_true)
        while len(bootstrapped_scores) < self.n_resamples:
            indices = self.random_state.choice(n_samples, n_samples, replace=True)
            try:
                score = metric_fn(y_true[indices], y_pred[indices], **kwargs)
                bootstrapped_scores.append(score)
            except:
                continue
                
        return bootstrapped_scores

    def bootstrap_ci(self, metric_fn: types.FunctionType, y_true: ArrayLike, y_pred: ArrayLike, alpha: float = 0.05, **kwargs) -> tuple:
        bootstrapped_scores = self.bootstrap(metric_fn = metric_fn, y_true = y_true, y_pred = y_pred, **kwargs)
        lower = np.percentile(bootstrapped_scores, 100 * (alpha / 2))
        upper = np.percentile(bootstrapped_scores, 100 * (1 - alpha / 2))
        return np.mean(bootstrapped_scores), lower, upper



from sklearn.preprocessing import label_binarize
import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from .metrics import BinaryClassificationMetrics, MulticlassClassificationMetrics


class FairnessMetrics:
    def __init__(
        self,
        y: ArrayLike,
        y_prob: ArrayLike,
        subgroup_indices: dict,
        threshold: float = 0.5,
        average: str = "macro",
    ):
        self.y_true = y
        self.y_prob = y_prob
        self.subgroup_indices = subgroup_indices
        self.threshold = threshold
        self.average = average
        self.n_classes = len(np.unique(y))

    def compute_fairness_metrics(self, bootstrap: bool = False, n_resamples: int = 1000):
        results = []
        gap_dict = {}

        # Decide metric class
        multiclass = self.n_classes > 2

        for feature, groups in self.subgroup_indices.items():
            metric_collect = {"TPR": [], "PPV": [], "FPR": []}  # for gap calculation

            for group_name, indices in groups.items():
                if len(indices) == 0:
                    continue

                y_g = self.y_true[indices]
                y_prob_g = self.y_prob[indices]

                if multiclass:
                    metric_obj = MulticlassClassificationMetrics(
                        y_g, y_prob_g, average=self.average
                    )
                    res = metric_obj.compute(bootstrap=bootstrap, n_resamples=n_resamples)

                    results.append({
                        "feature": feature,
                        "group": group_name,
                        "n": len(indices),
                        "TPR": res.get(f"recall ({self.average})"),
                        "TPR_lower": res.get(f"recall ({self.average})_lower"),
                        "TPR_upper": res.get(f"recall ({self.average})_upper"),
                        "PPV": res.get(f"precision ({self.average})"),
                        "PPV_lower": res.get(f"precision ({self.average})_lower"),
                        "PPV_upper": res.get(f"precision ({self.average})_upper"),
                        "FPR": np.nan  # Optional: not straightforward to define FPR for multiclass
                    })

                    metric_collect["TPR"].append(res.get(f"recall ({self.average})"))
                    metric_collect["PPV"].append(res.get(f"precision ({self.average})"))

                else:
                    y_prob_g_bin = y_prob_g[:, 1] if y_prob_g.ndim > 1 else y_prob_g
                    metric_obj = BinaryClassificationMetrics(
                        y_true=y_g,
                        y_pred_prob=y_prob_g_bin,
                        threshold=self.threshold
                    )
                    res = metric_obj.compute(bootstrap=bootstrap, n_resamples=n_resamples)

                    tpr = res["sensitivity"]
                    tpr_lower = res.get("sensitivity_lower")
                    tpr_upper = res.get("sensitivity_upper")
                    fpr = 1 - res["specificity"]
                    fpr_lower = 1 - res.get("specificity_upper") if "specificity_upper" in res else None
                    fpr_upper = 1 - res.get("specificity_lower") if "specificity_lower" in res else None
                    ppv = res["ppv"]
                    ppv_lower = res.get("ppv_lower")
                    ppv_upper = res.get("ppv_upper")

                    results.append({
                        "feature": feature,
                        "group": group_name,
                        "n": len(indices),
                        "TPR": tpr,
                        "TPR_lower": tpr_lower,
                        "TPR_upper": tpr_upper,
                        "FPR": fpr,
                        "FPR_lower": fpr_lower,
                        "FPR_upper": fpr_upper,
                        "PPV": ppv,
                        "PPV_lower": ppv_lower,
                        "PPV_upper": ppv_upper
                    })

                    metric_collect["TPR"].append(tpr)
                    metric_collect["FPR"].append(fpr)
                    metric_collect["PPV"].append(ppv)

            gap_dict[feature] = {
                "TPR_gap": np.nanmax(metric_collect["TPR"]) - np.nanmin(metric_collect["TPR"]),
                "PPV_gap": np.nanmax(metric_collect["PPV"]) - np.nanmin(metric_collect["PPV"]),
                "FPR_gap": (
                    np.nanmax(metric_collect["FPR"]) - np.nanmin(metric_collect["FPR"])
                    if not multiclass else np.nan
                ),
            }

        metrics_df = pd.DataFrame(results)
        gap_df = pd.DataFrame([
            {"feature": feature, **gaps} for feature, gaps in gap_dict.items()
        ])

        self.metrics_df = metrics_df
        self.gap_df = gap_df

        return metrics_df, gap_df


import numpy as np
from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from medevalkit.calibration import BinaryCalibration, MulticlassCalibration


def test_binary_calibration_returns_curve_and_brier_score():
    rng = np.random.RandomState(0)
    y_true = rng.randint(0, 2, size=200)
    logits = rng.normal(size=200)
    y_prob = 1 / (1 + np.exp(-logits))

    calib = BinaryCalibration(y_true, y_prob, n_bins=5)
    metrics = calib.compute(bootstrap=False)

    assert "brier_score" in metrics
    prob_pred, prob_true = metrics["calibration_curve"]
    assert prob_pred.shape == prob_true.shape


def test_multiclass_calibration_outputs_expected_structures():
    X, y = load_iris(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=42)
    model = RandomForestClassifier(n_estimators=50, random_state=42)
    model.fit(X_train, y_train)
    y_prob = model.predict_proba(X_test)

    mc = MulticlassCalibration(y_test, y_prob)
    ece = mc.expected_calibration_error(bootstrap=False)
    curves = mc.one_vs_rest_curves(bootstrap=False)

    assert "expected_calibration_error" in ece
    assert isinstance(curves, dict)
    for result in curves.values():
        assert "brier_score" in result
        curve = result["calibration_curve"]
        assert set(curve.keys()) == {"prob_pred", "prob_true"}



"""Threshold optimization strategies for classification models."""

from typing import Dict
from numpy.typing import ArrayLike
import numpy as np
import pandas as pd
from sklearn.metrics import precision_recall_curve, roc_curve, f1_score
from sklearn.preprocessing import label_binarize
from .metrics import BinaryClassificationMetrics

class ThresholdAnalysis:
    def __init__(self, y_true: ArrayLike, y_prob: ArrayLike):
        self.y_true = y_true
        self.y_pred_prob = y_prob

    def compute(self, step: float = 0.01) -> dict:
        output_dict = {}
        threshold = 1.0
        while threshold >= 0:
            cm = BinaryClassificationMetrics(y_true=self.y_true, y_pred_prob=self.y_pred_prob, threshold=threshold)
            output_dict[threshold] = cm.compute_static(bootstrap=False) if hasattr(cm, 'compute_static') else cm.compute()
            threshold -= step
        return output_dict

class DecisionCurveAnalysis:
    def __init__(self, y_true: ArrayLike, y_prob: ArrayLike, thresholds: ArrayLike = np.linspace(0.01, 0.99, 99)):
        self.y_true = y_true
        self.y_prob = y_prob
        self.thresholds = thresholds

    def compute(self) -> pd.DataFrame:
        results = []
        n = len(self.y_true)
        for thresh in self.thresholds:
            tp = np.sum((self.y_prob >= thresh) & (self.y_true == 1))
            fp = np.sum((self.y_prob >= thresh) & (self.y_true == 0))
            net_benefit = (tp / n) - (fp / n) * (thresh / (1 - thresh))
            results.append({'threshold': thresh, 'net_benefit': net_benefit})
        return pd.DataFrame(results)

class ThresholdOptimizer:
    def __init__(self, y_true: ArrayLike, y_prob: ArrayLike):
        self.y_true = y_true
        self.y_pred_prob = y_prob

    def optimize_youden(self) -> float:
        fpr, tpr, thresholds = roc_curve(self.y_true, self.y_pred_prob)
        j_scores = tpr - fpr
        return thresholds[np.argmax(j_scores)]

    def optimize_f1(self) -> float:
        precision, recall, thresholds = precision_recall_curve(self.y_true, self.y_pred_prob)
        f1_scores = 2 * (precision * recall) / (precision + recall + 1e-8)
        return thresholds[np.argmax(f1_scores)]

    def optimize_closest_to_perfect(self) -> float:
        fpr, tpr, thresholds = roc_curve(self.y_true, self.y_pred_prob)
        dist = np.sqrt((1 - tpr) ** 2 + fpr ** 2)
        return thresholds[np.argmin(dist)]

    def optimize_fixed_sensitivity(self, target_sens=0.9) -> float:
        fpr, tpr, thresholds = roc_curve(self.y_true, self.y_pred_prob)
        idx = np.where(tpr >= target_sens)[0]
        return thresholds[idx[0]] if len(idx) > 0 else thresholds[-1]

    def optimize_decision_curve(self) -> float:
        dca = DecisionCurveAnalysis(self.y_true, self.y_pred_prob)
        dca_df = dca.compute()
        return dca_df.loc[dca_df['net_benefit'].idxmax(), 'threshold']

class MulticlassThresholdOptimizer:
    """Threshold optimization for multiclass classifiers using one-vs-rest approach.

    Each class is treated as a separate binary classification problem, and
    thresholds are optimized independently using specified strategy.

    Attributes:
        y_true (np.ndarray): True labels (integer-encoded).
        y_pred_prob (np.ndarray): Predicted probabilities, shape (n_samples, n_classes).
        classes (np.ndarray): Unique class labels.
        thresholds_ (dict): Mapping from class label to optimal threshold.
    """

    def __init__(self, y_true: ArrayLike, y_prob: ArrayLike):
        """Initializes the optimizer.

        Args:
            y_true (ArrayLike): True labels (integers, shape (n_samples,)).
            y_pred_prob (ArrayLike): Predicted probabilities (shape (n_samples, n_classes)).
        """
        self.y_true = np.asarray(y_true)
        self.y_pred_prob = np.asarray(y_prob)
        self.classes = np.unique(self.y_true)
        self.n_classes = len(self.classes)
        self.thresholds_: Dict[int, float] = {}

        assert self.y_pred_prob.shape[1] == self.n_classes, \
            "Shape mismatch: y_pred_prob must have shape (n_samples, n_classes)"

    def optimize_per_class(self, method: str = "youden", **kwargs) -> Dict[int, float]:
        """Optimizes thresholds for each class using a specified method.

        Args:
            method (str): Optimization strategy. One of:
                - "youden"
                - "f1"
                - "closest_to_perfect"
                - "fixed_sensitivity"
                - "decision_curve"
            **kwargs: Additional keyword arguments for the optimization method.

        Returns:
            Dict[int, float]: Dictionary mapping class label to threshold.
        """
        self.thresholds_ = {}

        for class_index, class_label in enumerate(self.classes):
            # One-vs-rest binarization
            y_bin = (self.y_true == class_label).astype(int)
            y_prob = self.y_pred_prob[:, class_index]

            # Use the binary optimizer
            opt = ThresholdOptimizer(y_bin, y_prob)

            if method == "youden":
                threshold = opt.optimize_youden()
            elif method == "f1":
                threshold = opt.optimize_f1()
            elif method == "closest_to_perfect":
                threshold = opt.optimize_closest_to_perfect()
            elif method == "fixed_sensitivity":
                threshold = opt.optimize_fixed_sensitivity(**kwargs)
            elif method == "decision_curve":
                threshold = opt.optimize_decision_curve()
            else:
                raise ValueError(f"Unknown method: {method}")

            self.thresholds_[class_label] = threshold

        return self.thresholds_

    def predict(self, y_pred_prob: ArrayLike = None) -> np.ndarray:
        """Generates multiclass predictions using optimized thresholds.

        Args:
            y_pred_prob (ArrayLike, optional): Predicted probabilities to use.
                If None, will use the instance's stored probabilities.

        Returns:
            np.ndarray: Array of predicted class labels.
        """
        if not self.thresholds_:
            raise RuntimeError("Thresholds not computed. Call optimize_per_class() first.")

        y_pred_prob = self.y_pred_prob if y_pred_prob is None else np.asarray(y_pred_prob)
        decisions = np.full(y_pred_prob.shape[0], fill_value=-1, dtype=int)

        for i in range(y_pred_prob.shape[0]):
            passed = [
                cls for cls in self.classes
                if y_pred_prob[i, cls] >= self.thresholds_[cls]
            ]
            decisions[i] = passed[0] if passed else np.argmax(y_pred_prob[i])

        return decisions


# Re-importing necessary libraries after code execution environment reset
import numpy as np
import pandas as pd
import pandas as pd
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt


def compute_sens_spec_at_thresholds(results_df, pos_label=1, target_specs=[0.95], target_senss=[0.95]):
    """
    Compute sensitivity at given specificity thresholds, and vice versa.

    Parameters:
        results_df (pd.DataFrame): Contains 'pred_prob' and 'y_true'
        pos_label (int): Positive class label
        target_specs (List[float]): List of target specificity points
        target_senss (List[float]): List of target sensitivity points

    Returns:
        pd.DataFrame: Table of sensitivity at specificity and specificity at sensitivity
    """
    thresholds = np.linspace(0, 1, 1000)
    records = []

    for thresh in thresholds:
        preds = (results_df['pred_prob'] >= thresh).astype(int)
        tn, fp, fn, tp = confusion_matrix(results_df['y_true'], preds, labels=[0, 1]).ravel()

        sens = tp / (tp + fn) if (tp + fn) > 0 else 0
        spec = tn / (tn + fp) if (tn + fp) > 0 else 0

        records.append((thresh, sens, spec))

    metrics_df = pd.DataFrame(records, columns=['threshold', 'sensitivity', 'specificity'])

    results = []

    # Sensitivity at given specificity
    for spec_target in target_specs:
        subset = metrics_df[metrics_df['specificity'] >= spec_target]
        if not subset.empty:
            best_row = subset.sort_values('specificity').head(1).iloc[0]
            results.append({
                'metric': f'sensitivity_at_{int(spec_target*100)}%_specificity',
                'value': best_row['sensitivity'],
                'threshold': best_row['threshold']
            })

    # Specificity at given sensitivity
    for sens_target in target_senss:
        subset = metrics_df[metrics_df['sensitivity'] >= sens_target]
        if not subset.empty:
            best_row = subset.sort_values('sensitivity').head(1).iloc[0]
            results.append({
                'metric': f'specificity_at_{int(sens_target*100)}%_sensitivity',
                'value': best_row['specificity'],
                'threshold': best_row['threshold']
            })

    return pd.DataFrame(results)

def plot_trade_off(result_df):
    """
    Plot sensitivity at specificity and specificity at sensitivity trade-offs.

    Parameters:
        result_df (pd.DataFrame): DataFrame containing 'metric', 'value', and 'threshold'
    """
    # Plotting with deep blue color
    plt.figure(figsize=(10, 6))
    bars = plt.barh(result_df['metric'], result_df['value'], color='darkblue')
    plt.xlabel('Score')
    plt.title('Sensitivity at Specificity and Specificity at Sensitivity (Updated Thresholds - Blue)')
    plt.grid(axis='x', linestyle='--', alpha=0.7)

    # Annotate bars with value and threshold
    for i, bar in enumerate(bars):
        score = result_df['value'][i]
        threshold = result_df['threshold'][i]
        plt.text(bar.get_width() + 0.01, bar.get_y() + bar.get_height() / 2,
                    f"{score:.2f} @ {threshold:.2f}", va='center', color='black')

    plt.tight_layout()
    plt.show()

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--result_save_path', type=str, default="DeepECG/performance_test/saved_results/1_year_ecg_meta_td_all_lead.csv")
    args = parser.parse_args()

    results_df = pd.read_csv(args.result_save_path)
    result_table = compute_sens_spec_at_thresholds(
        results_df, target_specs=[0.9, 0.95, 0.99], target_senss=[0.9, 0.95, 0.99]
    )
    plot_trade_off(result_table)




import numpy as np
import pytest
from sklearn.datasets import make_classification, make_regression
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.model_selection import train_test_split

from medevalkit.evaluate import Evaluate


def test_evaluate_binary_classification_report_contains_metrics():
    X, y = make_classification(
        n_samples=400,
        n_features=10,
        n_informative=6,
        n_redundant=0,
        random_state=42,
    )
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=0)

    model = LogisticRegression(max_iter=1000)
    model.fit(X_train, y_train)
    y_prob = model.predict_proba(X_test)

    evaluator = Evaluate(y_true=y_test, y_prob=y_prob, classification=True, threshold=0.5)
    report = evaluator.generate_report(bootstrap=False)

    assert "clf_metrics" in report
    assert "calib_metrics" in report
    assert "text_report" in report
    assert report["clf_metrics"]["accuracy"] > 0.6
    assert "brier_score" in report["calib_metrics"]


def test_evaluate_multiclass_classification_support():
    X, y = make_classification(
        n_samples=300,
        n_features=12,
        n_classes=3,
        n_informative=6,
        n_redundant=0,
        n_clusters_per_class=1,
        random_state=21,
    )
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=1)

    model = LogisticRegression(max_iter=1000, multi_class="multinomial")
    model.fit(X_train, y_train)
    y_prob = model.predict_proba(X_test)

    evaluator = Evaluate(y_true=y_test, y_prob=y_prob, classification=True)
    report = evaluator.generate_report(bootstrap=False)

    clf_metrics = report["clf_metrics"]
    assert "auc (macro)" in clf_metrics
    assert "precision (macro)" in clf_metrics
    assert "recall (macro)" in clf_metrics
    assert report["text_report"]


def test_evaluate_regression_report_contains_metrics():
    X, y = make_regression(n_samples=200, n_features=8, noise=0.5, random_state=123)
    X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=12)

    model = LinearRegression()
    model.fit(X_train, y_train)
    y_pred = model.predict(X_test)

    evaluator = Evaluate(y_true=y_test, y_pred=y_pred, classification=False)
    report = evaluator.generate_report(bootstrap=False)

    assert "reg_metrics" in report
    assert "mse" in report["reg_metrics"]
    assert "text_report" in report


def test_evaluate_raises_when_probabilities_missing_for_classification():
    y_true = np.array([0, 1, 0, 1])
    with pytest.raises(ValueError) as excinfo:
        Evaluate(y_true=y_true, classification=True)
    assert "y_prob" in str(excinfo.value)



from typing import Dict, List, Tuple
import matplotlib.pyplot as plt
from lifelines import KaplanMeierFitter
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.metrics import roc_curve, auc, roc_auc_score
from sklearn.calibration import calibration_curve
from sklearn.preprocessing import label_binarize
from .compare import ModelComparer, delong_test
from .bootstrap import Bootstrapper

# Set custom style sheet for all plots
plt.style.use('medevalkit/styles/ieee.mplstyle')

def plot_roc_curve(y_true, y_prob, label='Model', save_path=None):
    """
    Plots a Receiver Operating Characteristic (ROC) curve.

    Args:
        y_true (array-like): True binary labels.
        y_prob (array-like): Predicted probabilities for the positive class.
        label (str): Label for the model in the legend.
        save_path (str, optional): Path to save the plot (e.g., 'roc_curve.png').
    """
    fpr, tpr, _ = roc_curve(y_true, y_prob)
    roc_auc = auc(fpr, tpr)

    plt.plot(fpr, tpr, label=f'{label} (AUC = {roc_auc:.3f})')
    plt.plot([0, 1], [0, 1], color='gray', linestyle='--', label='Random Guess')

    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver Operating Characteristic (ROC) Curve')
    plt.legend()
    plt.xlim([-0.01, 1.01])
    plt.ylim([-0.01, 1.01])
    plt.gca().set_aspect('equal', adjustable='box') # Ensure 1:1 aspect ratio
    if save_path:
        plt.savefig(save_path)
    plt.show()

def plot_multiclass_roc(y_true, y_score, class_labels=None, save_path=None):
    """
    Plot ROC curves for multiclass classification using One-vs-Rest.

    Args:
        y_true (array-like): True class labels (n_samples,).
        y_score (array-like): Predicted probabilities (n_samples, n_classes).
        class_labels (list, optional): List of class names.
        save_path (str, optional): Path to save the plot.
    """
    n_classes = y_score.shape[1]
    if class_labels is None:
        class_labels = [f'Class {i}' for i in range(n_classes)]

    y_true_bin = label_binarize(y_true, classes=np.arange(n_classes))

    for i in range(n_classes):
        fpr, tpr, _ = roc_curve(y_true_bin[:, i], y_score[:, i])
        roc_auc = auc(fpr, tpr)
        plt.plot(fpr, tpr, label=f'{class_labels[i]} (AUC = {roc_auc:.3f})')

    plt.plot([0, 1], [0, 1], 'k--', label='Random Guess')
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Multiclass ROC Curve (One-vs-Rest)')
    plt.legend()
    plt.grid(True)
    plt.xlim([-0.01, 1.01])
    plt.ylim([-0.01, 1.01])
    plt.gca().set_aspect('equal', adjustable='box')
    if save_path:
        plt.savefig(save_path)
    plt.show()

def plot_calibration_curve(y_true, y_prob, n_bins=10, label='Model', save_path=None):
    """
    Plots a calibration curve (reliability diagram).

    Args:
        y_true (array-like): True binary labels.
        y_prob (array-like): Predicted probabilities for the positive class.
        n_bins (int): Number of bins to use for calibration.
        label (str): Label for the model in the legend.
        save_path (str, optional): Path to save the plot.
    """
    prob_true, prob_pred = calibration_curve(y_true, y_prob, n_bins=n_bins)

    plt.plot([0, 1], [0, 1], linestyle='--', color='gray', label='Perfectly Calibrated')
    plt.plot(prob_pred, prob_true, marker='o', label=label)
    
    plt.xlabel('Mean Predicted Probability')
    plt.ylabel('Fraction of Positives (True Probability)')
    plt.title('Calibration Curve')
    plt.legend()
    plt.grid(True)
    plt.xlim([-0.01, 1.01])
    plt.ylim([-0.01, 1.01])
    plt.gca().set_aspect('equal', adjustable='box')
    if save_path:
        plt.savefig(save_path)
    plt.show()

def plot_multiclass_calibration(calibration_obj, class_labels=None, save_path=None):
    """
    Plots one-vs-rest calibration curves from a MulticlassCalibration instance.

    Args:
        calibration_obj: An instance of MulticlassCalibration (assumed to have one_vs_rest_curves method).
        class_labels (list, optional): Optional list of class names.
        save_path (str, optional): Path to save the plot.
    """
    curves = calibration_obj.one_vs_rest_curves()
    n_classes = len(curves)
    
    if class_labels is None:
        class_labels = [f'Class {i}' for i in range(n_classes)]

    plt.plot([0, 1], [0, 1], linestyle='--', color='gray', label='Perfectly Calibrated', alpha=0.7)

    for k, result in curves.items():
        prob_pred = result["calibration_curve"]["prob_pred"]
        prob_true = result["calibration_curve"]["prob_true"]
        brier = result["brier_score"]
        label = f"{class_labels[k]} (Brier: {brier:.3f})" # Format Brier score
        plt.plot(prob_pred, prob_true, marker="o", label=label)

    plt.xlabel("Mean Predicted Probability")
    plt.ylabel("Fraction of Positives")
    plt.title("Multiclass Calibration Curves (One-vs-Rest)")
    plt.legend()
    plt.grid(True)
    plt.xlim([-0.01, 1.01])
    plt.ylim([-0.01, 1.01])
    plt.gca().set_aspect('equal', adjustable='box')
    plt.tight_layout() # Adjust for external legend
    if save_path:
        plt.savefig(save_path)
    plt.show()

def plot_decision_curve(dca_df, label='Model', save_path=None):
    """
    Plots a Decision Curve Analysis (DCA) curve.

    Args:
        dca_df (pd.DataFrame): DataFrame with 'threshold' and 'net_benefit' columns.
        label (str): Label for the model in the legend.
        save_path (str, optional): Path to save the plot.
    """
    plt.plot(dca_df['threshold'], dca_df['net_benefit'], label=label)
    plt.axhline(0, linestyle='--', color='gray', label='Treat All/Treat None')
    
    # Optional: Plot 'treat all' and 'treat none' if available in dca_df or calculable
    if 'net_benefit_all' in dca_df.columns:
        plt.plot(dca_df['threshold'], dca_df['net_benefit_all'], label='Treat All')
    if 'net_benefit_none' in dca_df.columns:
        plt.plot(dca_df['threshold'], dca_df['net_benefit_none'], label='Treat None')

    plt.xlabel('Threshold Probability')
    plt.ylabel('Net Benefit')
    plt.title('Decision Curve Analysis')
    plt.legend()
    plt.grid(True)
    plt.xlim([0, 1])
    # Adjust y-axis limits to better fit data and include 0
    min_nb = dca_df['net_benefit'].min()
    max_nb = dca_df['net_benefit'].max()
    plt.ylim(min(0, min_nb - 0.05), max(0.1, max_nb + 0.05)) # Extend limits slightly
    
    plt.tight_layout()
    if save_path:
        plt.savefig(save_path)
    plt.show()

def plot_threshold_metrics(threshold_dict, metrics=('sensitivity', 'specificity', 'ppv', 'npv'), save_path=None):
    """
    Plots various classification metrics across different probability thresholds.

    Args:
        threshold_dict (dict): Dictionary with thresholds as keys and a dict of metrics as values.
        metrics (tuple): Metrics to plot (e.g., 'sensitivity', 'specificity').
        save_path (str, optional): Path to save the plot.
    """
    df = pd.DataFrame.from_dict(threshold_dict, orient='index').sort_index(ascending=True)
    
    for i, metric in enumerate(metrics):
        if metric in df.columns:
            plt.plot(df.index, df[metric], label=metric.replace('_', ' ').title())
            
    plt.xlabel('Threshold')
    plt.ylabel('Metric Value')
    plt.title('Threshold Analysis Metrics')
    plt.legend()
    plt.grid(True)
    plt.xlim([-0.01, 1.01])
    plt.ylim([-0.01, 1.01])
    plt.tight_layout() # Adjust for external legend
    if save_path:
        plt.savefig(save_path)
    plt.show()

def plot_regression_diagnostics(y_true, y_pred, save_path=None):
    """
    Plots regression diagnostics:
    - Predicted vs Actual values
    - Residuals vs Predicted values

    Args:
        y_true (array-like): True continuous targets.
        y_pred (array-like): Predicted continuous outputs.
        save_path (str, optional): Path to save the plot.
    """
    residuals = y_true - y_pred

    fig, axs = plt.subplots(1, 2, figsize=(14, 6)) # Larger figure

    # Predicted vs Actual
    axs[0].scatter(y_true, y_pred, alpha=0.7, s=50, edgecolors='w', linewidths=0.5)
    min_val = min(y_true.min(), y_pred.min())
    max_val = max(y_true.max(), y_pred.max())
    axs[0].plot([min_val, max_val], [min_val, max_val], 'r--', lw=2, label='Ideal Fit')
    axs[0].set_xlabel('Actual Values', fontsize=13)
    axs[0].set_ylabel('Predicted Values', fontsize=13)
    axs[0].set_title('Predicted vs Actual Values', fontsize=15)
    axs[0].grid(True, linestyle=':', alpha=0.7)
    axs[0].set_aspect('equal', adjustable='box')
    axs[0].legend(loc='upper left', fontsize=11)


    # Residuals vs Predicted
    axs[1].scatter(y_pred, residuals, alpha=0.7, s=50, edgecolors='w', linewidths=0.5)
    axs[1].axhline(0, linestyle='--', color='gray', lw=2, label='Zero Residuals')
    axs[1].set_xlabel('Predicted Values', fontsize=13)
    axs[1].set_ylabel('Residuals', fontsize=13)
    axs[1].set_title('Residuals vs Predicted Values', fontsize=15)
    axs[1].grid(True, linestyle=':', alpha=0.7)
    axs[1].legend(loc='upper right', fontsize=11)


    plt.tight_layout()
    if save_path:
        plt.savefig(save_path, dpi=300, bbox_inches='tight')
    plt.show()

def plot_multiple_roc_curves_with_comparison(
    y_true: np.ndarray,
    model_probs: Dict[str, np.ndarray],
    method: str = "delong",
    parametric: bool = True,
    n_resamples: int = 1000,
    return_stats: bool = False,
    show_pvalues_on_plot: bool = True,
    save_path: str = None
):
    """
    Plot multiple ROC curves and optionally annotate pairwise p-values.

    Args:
        y_true (np.ndarray): Ground truth binary labels (1D array)
        model_probs (Dict[str, np.ndarray]): Dict of model name -> predicted probabilities (1D array)
        method (str): 'delong' or 'bootstrap' for p-value calculation.
        parametric (bool): If using bootstrap, use parametric t-test (True) or Mann-Whitney U test (False).
        n_resamples (int): Number of bootstrap resamples.
        return_stats (bool): If True, return a DataFrame of p-values and AUCs.
        show_pvalues_on_plot (bool): If True, add pairwise p-values as text annotations on the plot.
        save_path (str, optional): Path to save the plot.
    """
    model_names = list(model_probs.keys())
    stats = []
    for i, name in enumerate(model_names):
        fpr, tpr, _ = roc_curve(y_true, model_probs[name])
        auc_val = auc(fpr, tpr)
        plt.plot(
            fpr, tpr,
            label=f"{name} (AUC = {auc_val:.3f})",
        )

    # Plot settings
    plt.plot(
        [0, 1], [0, 1],
        linestyle='--',
        label='Random Guess',
        color='gray'
    )
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.title("Comparison of ROC Curves")
    plt.grid(True)
    plt.minorticks_on()
    plt.grid(which='minor')
    plt.xlim(-0.01, 1.01)
    plt.ylim(-0.01, 1.01)
    plt.gca().set_aspect('equal', adjustable='box')

    # Pairwise p-value annotations outside the plot
    if show_pvalues_on_plot and len(model_names) > 1:
        y0 = 0.95
        for i in range(len(model_names)):
            for j in range(i + 1, len(model_names)):
                name1, name2 = model_names[i], model_names[j]
                if method == "delong":
                    pval, _, _ = delong_test(y_true, model_probs[name1], model_probs[name2])
                else:
                    bs = Bootstrapper(n_resamples=n_resamples)
                    dist1 = bs.bootstrap(roc_auc_score, y_true, model_probs[name1])
                    dist2 = bs.bootstrap(roc_auc_score, y_true, model_probs[name2])
                    test_fn = ModelComparer.parametric if parametric else ModelComparer.nonparametric
                    pval = test_fn(dist1, dist2)

                stats.append({"Model 1": name1, "Model 2": name2, "p-value": pval})
                p_str = 'p < 0.001' if pval < 0.001 else f'p = {pval:.3f}'

                plt.annotate(
                    f"{name1} vs {name2}: {p_str}",
                    xy=(1.02, y0), xycoords='axes fraction',
                    ha='left', va='top', fontsize=8
                )
                y0 -= 0.05

    plt.legend()

    if save_path:
        plt.savefig(save_path)
    plt.show()

    if return_stats:
        return pd.DataFrame(stats)

def plot_auc_comparison_heatmap(
    comparison_df: pd.DataFrame,
    metric: str = "p-value",
    cmap: str = "RdYlBu_r", # Diverging colormap
    annot: bool = True,
    fmt: str = ".3f",
    figsize: tuple = None,
    title: str = None,
    save_path: str = None
):
    """
    Plot a symmetric heatmap from a pairwise AUC comparison dataframe.

    Args:
        comparison_df (pd.DataFrame): Output DataFrame with columns: ["Model 1", "Model 2", metric]
        metric (str): Column name to visualize (e.g., "p-value", "AUC 1", "AUC 2").
        cmap (str): Colormap for the heatmap.
        annot (bool): Annotate heatmap cells.
        fmt (str): Format for annotation values.
        figsize (tuple, optional): Optional custom figure size.
        title (str, optional): Optional title (default: auto-generated).
        save_path (str, optional): Path to save the plot.
    """
    if not {"Model 1", "Model 2", metric}.issubset(comparison_df.columns):
        raise ValueError(f"Input DataFrame must contain 'Model 1', 'Model 2', and '{metric}' columns.")

    models = sorted(set(comparison_df["Model 1"]).union(set(comparison_df["Model 2"])))
    matrix = pd.DataFrame(index=models, columns=models, dtype=object) # Use object to store formatted strings

    for _, row in comparison_df.iterrows():
        m1, m2 = row["Model 1"], row["Model 2"]
        value = row[metric]
        
        display_value = ""
        if metric == 'p-value':
            if value < 0.001:
                display_value = '< 0.001'
            else:
                display_value = f'{value:.3f}'
        else:
            display_value = f'{value:{fmt}}' # Apply general format
        
        matrix.loc[m1, m2] = display_value
        matrix.loc[m2, m1] = display_value

    # Fill diagonal with empty string or model name if desired
    for model in models:
        matrix.loc[model, model] = "" # Or model name, e.g., f"AUC: {aucs.get(model, 'N/A'):.3f}" if AUCs are available

    if figsize is None:
        n = len(models)
        figsize = (max(8, n * 1.2), max(7, n)) # Adjusted size for readability

    if title is None:
        title = f"Pairwise {metric} Comparison between Models"

    plt.figure(figsize=figsize)
    # Convert matrix to numeric for cmap if metric is numeric, else cmap might not apply well
    # For p-values, we might want a custom colormap that highlights significance
    numeric_matrix = matrix.apply(pd.to_numeric, errors='coerce') # For cmap
    
    sns.heatmap(numeric_matrix, annot=matrix, fmt="", cmap=cmap, square=True,
                linewidths=0.5, linecolor='black', cbar_kws={"label": metric.replace('_', ' ').title()})
    plt.title(title, fontsize=16)
    plt.yticks(rotation=0)
    plt.xticks(rotation=45, ha="right")
    plt.tight_layout()
    if save_path:
        plt.savefig(save_path, dpi=300, bbox_inches='tight')
    plt.show()

def plot_auc_bar_chart_with_error_bars(
    auc_dict: Dict[str, np.ndarray],
    pval_df: pd.DataFrame = None, # Changed to DataFrame for more flexible p-value display
    ci: float = 0.95,
    title: str = "AUC Comparison with Confidence Intervals",
    annotate: bool = True,
    ylim: tuple = None,
    save_path: str = None
):
    """
    Plots a bar chart of AUCs with confidence intervals and optional p-value annotations.

    Args:
        auc_dict (Dict[str, np.ndarray]): Dictionary of model name -> bootstrap distribution of AUCs.
        pval_df (pd.DataFrame, optional): DataFrame with pairwise p-values (from plot_multiple_roc_curves_with_comparison).
                                        Expected columns: "Model 1", "Model 2", "p-value".
        ci (float): Confidence interval level (e.g., 0.95 for 95% CI).
        title (str): Plot title.
        figsize (tuple): Figure size.
        annotate (bool): If True, annotate bars with AUC and CI.
        ylim (tuple, optional): Custom y-axis limits.
        save_path (str, optional): Path to save the plot.
    """
    colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

    model_names = list(auc_dict.keys())
    medians = [np.median(auc_dict[m]) for m in model_names]
    alpha = (1 - ci) / 2 * 100
    lowers = [np.percentile(auc_dict[m], alpha) for m in model_names]
    uppers = [np.percentile(auc_dict[m], 100 - alpha) for m in model_names]
    errs = np.array([[m - lo, hi - m] for m, lo, hi in zip(medians, lowers, uppers)]).T

    fig, ax = plt.subplots()
    bars = ax.bar(
        model_names,
        medians,
        yerr=errs,
        capsize=3,
        color=colors[:len(model_names)]
    )

    # Spacing
    bar_label_offset    = 0.002 
    bracket_base_offset = 0.008    
    bracket_height      = 0.0015   
    pval_text_offset    = 0.001 
    bar_label_fs = 7
    pval_text_fs = 6

    if annotate:
        for bar, m, lo, hi in zip(bars, medians, lowers, uppers):
            ax.text(
                bar.get_x() + bar.get_width()/2,
                hi + bar_label_offset,
                f"{m:.3f}\n[{lo:.3f}, {hi:.3f}]",
                ha="center", va="bottom",
                fontsize=bar_label_fs
            )

    # Draw p-value brackets
    used_heights = {}
    if pval_df is not None and not pval_df.empty:
        pval_df = pval_df.sort_values("p-value").reset_index(drop=True)
        for _, row in pval_df.iterrows():
            m1, m2, p = row["Model 1"], row["Model 2"], row["p-value"]
            i1, i2 = model_names.index(m1), model_names.index(m2)
            x1 = bars[i1].get_x() + bars[i1].get_width()/2
            x2 = bars[i2].get_x() + bars[i2].get_width() + bars[i2].get_width()/2 - bars[i2].get_width()

            base = max(uppers[i1], uppers[i2]) + bracket_base_offset
            y0 = base + (bracket_height + 0.001) * len(used_heights)
            used_heights[m1] = used_heights[m2] = y0

            ax.plot(
                [x1, x1, x2, x2],
                [y0, y0 + bracket_height, y0 + bracket_height, y0],
                color="black"
            )

            p_str = "p < 0.001" if p < 0.001 else f"p = {p:.3f}"
            ax.text(
                (x1 + x2)/2,
                y0 + bracket_height + pval_text_offset,
                p_str,
                ha="center", va="bottom",
                fontsize=pval_text_fs
            )

    # Y-limits
    if ylim:
        ax.set_ylim(*ylim)
    else:
        low = min(lowers) - 0.005
        high_candidates = uppers + list(used_heights.values())
        high = max(high_candidates) + 0.005
        ax.set_ylim(low, high)

    ax.set_ylabel("AUC")
    ax.set_title(title)
    ax.grid(True)
    plt.xticks()

    if save_path:
        fig.savefig(save_path)
    plt.show()

def plot_kaplan_meier(durations, events, ax=None, label=None, save_path=None, **kwargs):
    """
    Plot Kaplan-Meier survival curve.

    Args:
        durations (array-like): Time to event or censoring.
        events (array-like): Event indicators (1=event, 0=censored).
        ax (matplotlib.axes.Axes, optional): Existing axis to plot into.
        label (str, optional): Label for the curve.
        save_path (str, optional): Path to save the plot.
        **kwargs: Additional arguments passed to lifelines.KaplanMeierFitter.plot_survival_function.

    Returns:
        ax: The matplotlib axes object.
    """

    if ax is None:
        fig, ax = plt.subplots(figsize=(8, 6)) # Create new figure if no ax provided
    else:
        fig = ax.figure # Get figure from ax

    kmf = KaplanMeierFitter()
    kmf.fit(durations, event_observed=events, label=label)
    
    # Customize plot_survival_function arguments for publication quality
    kmf.plot_survival_function(
        ax=ax, 
        ci_show=True,       # Show confidence intervals
        color='steelblue',  # Line color
        ls='-',             # Line style
        lw=2,               # Line width
        censor_styles={'marker': 'o', 'ms': 6, 'mew': 1.5, 'alpha': 0.7}, # Censor marker style
        **kwargs
    )

    ax.set_title("Kaplan-Meier Survival Curve", fontsize=16)
    ax.set_xlabel("Time", fontsize=14)
    ax.set_ylabel("Survival Probability", fontsize=14)
    ax.grid(True, linestyle=':', alpha=0.7)
    ax.legend(loc='lower left', fontsize=12) # Better legend placement for survival curves
    ax.set_ylim([-0.02, 1.02]) # Ensure y-axis covers 0 to 1 with slight padding
    ax.set_xlim(left=0) # Ensure x-axis starts at 0 or appropriate minimum

    plt.tight_layout()
    if save_path and ax is fig.gca(): # Only save if we created the figure in this function
        plt.savefig(save_path, dpi=300, bbox_inches='tight')
    elif save_path and fig: # If an ax was passed but we still want to save the figure it belongs to
        fig.savefig(save_path, dpi=300, bbox_inches='tight')
        
    if ax is fig.gca(): # Only show if we created the figure
        plt.show()
    return ax


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.utils import resample
from typing import Tuple, List, Optional, Dict
from sksurv.util import Surv
from sksurv.metrics import cumulative_dynamic_auc
from lifelines import KaplanMeierFitter


class LandmarkSurvivalEvaluator:
    """Evaluate a classifier using landmarking survival approach."""

    def __init__(self, model, prediction_window: float):
        """
        Args:
            model: A trained classifier with predict_proba method.
            prediction_window (float): Landmark time window for defining events.
        """
        self.model = model
        self.prediction_window = prediction_window
        self._train_X = None
        self._train_durations = None
        self._train_events = None

    def fit(self, X_train: np.ndarray, durations: np.ndarray, events: np.ndarray):
        """
        Store training survival data for dynamic AUC evaluation.

        Args:
            X_train: Feature matrix used to train the model.
            durations: Follow-up times.
            events: Event indicators (1=event, 0=censored).
        """
        self._train_X = np.asarray(X_train)
        self._train_durations = np.asarray(durations)
        self._train_events = np.asarray(events)

    def transform_labels(self, durations: np.ndarray, events: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Convert survival data into binary labels and a mask indicating valid samples.

        Args:
            durations: Follow-up times.
            events: Event indicators.

        Returns:
            label: Binary event within prediction window.
            mask: Whether subject is still at risk at prediction window.
        """
        durations = np.asarray(durations)
        events = np.asarray(events)
        label = (durations <= self.prediction_window) & (events == 1)
        mask = durations >= self.prediction_window
        return label.astype(int), mask

    def evaluate_dynamic_auc(self, X, durations, events, times: list, n_resamples=1000, seed=1):
        """
        Computes time-dependent AUCs via cumulative_dynamic_auc with test set bootstrapping.
        """
        if self._train_X is None:
            raise ValueError("Must call .fit() with training data before evaluating dynamic AUC.")

        train_surv = Surv.from_arrays(self._train_events == 1, self._train_durations)
        test_pred = self.model.predict_proba(X)[:, 1]
        rng = np.random.default_rng(seed)

        results = {t: [] for t in times}
        successful = 0

        for _ in range(n_resamples):
            try:
                # Bootstrap resample test set
                X_b, d_b, e_b, y_b = resample(X, durations, events, test_pred, random_state=rng.integers(1e9))
                if len(np.unique(y_b)) < 2:
                    continue  # skip if predictions lack variance

                surv_b = Surv.from_arrays(e_b == 1, d_b)
                aucs, _ = cumulative_dynamic_auc(train_surv, surv_b, y_b, times=times)

                for i, t in enumerate(times):
                    results[t].append(aucs[i])
                successful += 1
            except Exception:
                for t in times:
                    results[t].append(np.nan)

        if successful == 0:
            raise RuntimeError("All bootstrap iterations failed.")

        df = pd.DataFrame.from_dict(results, orient='index')
        summary = df.apply(lambda x: pd.Series({
            "mean": np.nanmean(x),
            "lower": np.nanpercentile(x, 2.5),
            "upper": np.nanpercentile(x, 97.5)
        }), axis=1)

        summary.index.name = "time"
        return summary.reset_index()

    def evaluate_multiple_windows(
        self,
        X: np.ndarray,
        durations: np.ndarray,
        events: np.ndarray,
        window_list: List[float],
        times: List[float],
        n_resamples: int = 1000
    ) -> Dict[float, pd.DataFrame]:
        """
        Evaluate dynamic AUCs across multiple prediction windows.

        Args:
            X: Test feature matrix.
            durations: Test follow-up times.
            events: Event indicators.
            window_list: List of prediction windows to evaluate.
            times: Time points for AUC evaluation.
            n_resamples: Bootstrap iterations.

        Returns:
            Dictionary of {window: AUC DataFrame}.
        """
        results = {}
        for w in window_list:
            self.prediction_window = w
            results[w] = self.evaluate_dynamic_auc(X, durations, events, times, n_resamples)
        return results

    def stratified_kaplan_by_risk(
        self,
        durations: np.ndarray,
        events: np.ndarray,
        probs: np.ndarray,
        n_bins: int = 4,
        ax: Optional[plt.Axes] = None
    ) -> plt.Axes:
        """
        Plot Kaplan-Meier survival curves stratified by predicted risk.

        Args:
            durations: Follow-up times.
            events: Event indicators.
            probs: Predicted probabilities.
            n_bins: Number of risk bins.
            ax: Optional matplotlib axis.

        Returns:
            The matplotlib axis with the plot.
        """
        df = pd.DataFrame({
            'risk_score': probs,
            'duration': durations,
            'event': events
        })
        df['bin'] = pd.qcut(df['risk_score'], n_bins, labels=False)

        if ax is None:
            _, ax = plt.subplots(figsize=(8, 6))

        for b in sorted(df['bin'].unique()):
            mask = df['bin'] == b
            kmf = KaplanMeierFitter()
            kmf.fit(df.loc[mask, 'duration'], df.loc[mask, 'event'])
            kmf.plot_survival_function(ax=ax, label=f'Q{b + 1}')

        ax.set_title("Kaplan-Meier Curves by Risk Score Quartile")
        ax.set_xlabel("Time")
        ax.set_ylabel("Survival Probability")
        ax.legend()
        ax.grid(True)
        return ax


import numpy as np
from sklearn.datasets import load_breast_cancer
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import roc_auc_score
from medevalkit.compare import ModelComparer, delong_test

def test_delong_test_statistic_validity():
    np.random.seed(42)
    y_true = np.random.randint(0, 2, size=100)
    y_score1 = np.random.rand(100)
    y_score2 = np.random.rand(100)

    p, auc1, auc2 = delong_test(y_true, y_score1, y_score2)
    assert 0 <= auc1 <= 1
    assert 0 <= auc2 <= 1
    assert 0 <= p <= 1

def test_model_comparer_bootstrap_and_delong():
    data = load_breast_cancer()
    X_train, X_test, y_train, y_test = train_test_split(data.data, data.target, stratify=data.target, random_state=42)
    clf1 = LogisticRegression(max_iter=1000).fit(X_train, y_train)
    clf2 = RandomForestClassifier(n_estimators=100).fit(X_train, y_train)

    comparer = ModelComparer({"LogReg": clf1, "RF": clf2}, X_test, y_test)

    result_bootstrap = comparer.compare_auc(method="bootstrap", parametric=True, n_resamples=100)
    result_delong = comparer.compare_auc(method="delong")

    assert isinstance(result_bootstrap, dict)
    assert isinstance(result_delong, dict)
    assert len(result_bootstrap) == len(result_delong)



"""Simulation tools for synthetic data or perturbation experiments."""

import math
import numpy as np
from numpy.typing import ArrayLike

class PopulationSimulator:
    """
    Create a set of virtual cohorts of a prespecified incidence rate using bootstrap resampling 
    """
    def __init__(self, y: ArrayLike, n_resamples: int, random_state=None):
        try:
            self.y = y.tolist()
        except:
            self.y = y

        labels = set(np.unique(y))
        if labels != {0, 1}:
            raise ValueError("Simulator only supports binary classification with labels 0 and 1.")
        self.incidence_rate = sum(y)/len(y)
        self.n_resamples = n_resamples
        self.random_state = np.random.RandomState(random_state)

    def simulate(self, incidence_rate_list: list = [0.01, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.99]) -> tuple:
        positive_cases = [idx for idx,y_label in enumerate(self.y) if y_label == 1]
        negative_cases = [idx for idx,y_label in enumerate(self.y) if y_label == 0]
        
        simulation_dict = {}
        incidence_rate_list = sorted(list(set(incidence_rate_list + [self.incidence_rate])))

        for i in incidence_rate_list:
            resampled_pos = []
            resampled_neg = []
            for _ in range(self.n_resamples):
                #randomly resample from y such that incidence matches i
                n_pos = max(math.ceil(i * len(self.y)),1)
                n_neg = max(math.ceil((1-i) * len(self.y)),1)
                resampled_pos += [self.random_state.choice(positive_cases, n_pos, replace=True)]
                resampled_neg += [self.random_state.choice(negative_cases, n_neg, replace=True)]
            #concat arrays
            resampled_pos = np.concatenate(resampled_pos).tolist()
            resampled_neg = np.concatenate(resampled_neg).tolist()
            #combined arrays
            combined = resampled_pos + resampled_neg
            #resample to correct length - n_resamples * len(y)
            combined = self.random_state.choice(combined, self.n_resamples * len(self.y), replace=True)
            assert len(combined) == self.n_resamples * len(self.y)
            assert len(combined) 
            simulation_dict[i] = combined

        return simulation_dict



import numpy as np
from sklearn.datasets import make_classification
from sklearn.linear_model import LogisticRegression
from sklearn.model_selection import train_test_split

from medevalkit.evaluate import EvaluateWithSimulation


def test_simulation_generates_summary_for_requested_incidence_rates():
    X, y = make_classification(
        n_samples=500,
        n_features=10,
        n_informative=6,
        weights=[0.6, 0.4],
        random_state=123,
    )
    X_train, X_test, y_train, y_test = train_test_split(X, y, stratify=y, random_state=7)

    clf = LogisticRegression(max_iter=1000)
    clf.fit(X_train, y_train)
    y_prob = clf.predict_proba(X_test)

    simulator = EvaluateWithSimulation(y_true=y_test, y_prob=y_prob, threshold=0.5)
    incidence_rates = [0.2, 0.5]
    results = simulator.run_simulation(
        incidence_rate_list=incidence_rates,
        calibration_bins=5,
        bootstrap=False,
        n_resamples=3,
        random_state=5,
    )

    for rate in incidence_rates:
        assert rate in results

    for summary in results.values():
        assert "auc" in summary or "accuracy" in summary

    metrics_df = simulator.generate_metrics()
    assert not metrics_df.empty
    assert any("auc" in idx for idx in metrics_df.index)



import numpy as np

from medevalkit.threshold import (
    DecisionCurveAnalysis,
    MulticlassThresholdOptimizer,
    ThresholdAnalysis,
    ThresholdOptimizer,
)


def test_threshold_analysis_returns_metrics_for_requested_steps():
    y_true = np.array([0, 0, 1, 1, 0, 1, 0, 1])
    y_prob = np.array([0.1, 0.4, 0.75, 0.8, 0.2, 0.85, 0.3, 0.7])

    analysis = ThresholdAnalysis(y_true, y_prob)
    summary = analysis.compute(step=0.5)

    assert set(summary.keys()) == {1.0, 0.5, 0.0}
    for metrics in summary.values():
        assert {"sensitivity", "specificity", "accuracy"}.issubset(metrics.keys())


def test_decision_curve_analysis_dataframe_structure():
    y_true = np.array([0, 1, 1, 0, 1, 0, 1, 0, 0, 1])
    y_prob = np.array([0.2, 0.8, 0.65, 0.3, 0.7, 0.1, 0.9, 0.4, 0.25, 0.6])

    dca = DecisionCurveAnalysis(y_true, y_prob)
    df = dca.compute()

    assert {"threshold", "net_benefit"}.issubset(df.columns)
    assert (df["threshold"] >= 0).all() and (df["threshold"] <= 1).all()


def test_threshold_optimizers_return_valid_thresholds():
    y_true = np.array([0, 0, 1, 1, 0, 1, 0, 1])
    y_prob = np.array([0.05, 0.35, 0.81, 0.76, 0.22, 0.88, 0.4, 0.7])

    optimizer = ThresholdOptimizer(y_true, y_prob)
    youden = optimizer.optimize_youden()
    f1 = optimizer.optimize_f1()

    assert 0 <= youden <= 1
    assert 0 <= f1 <= 1


def test_multiclass_threshold_optimizer_predicts_labels():
    y_true = np.array([0, 1, 2, 1, 0, 2])
    y_prob = np.array(
        [
            [0.7, 0.2, 0.1],
            [0.1, 0.8, 0.1],
            [0.1, 0.2, 0.7],
            [0.2, 0.6, 0.2],
            [0.8, 0.1, 0.1],
            [0.2, 0.2, 0.6],
        ]
    )

    mto = MulticlassThresholdOptimizer(y_true, y_prob)
    thresholds = mto.optimize_per_class(method="youden")
    predictions = mto.predict()

    assert thresholds.keys() == {0, 1, 2}
    assert predictions.shape == y_true.shape
    assert set(predictions).issubset({0, 1, 2})



import numpy as np

from medevalkit.metrics import (
    BinaryClassificationMetrics,
    MulticlassClassificationMetrics,
    RegressionMetrics,
)


def test_binary_classification_metrics_basic_values():
    y_true = np.array([0, 1, 1, 0, 1])
    y_prob = np.array([0.2, 0.9, 0.7, 0.1, 0.85])
    bcm = BinaryClassificationMetrics(y_true, y_prob, threshold=0.5)
    results = bcm.compute(bootstrap=False)

    expected_keys = {"auc", "auprc", "f1", "sensitivity", "specificity", "accuracy"}
    assert expected_keys.issubset(results.keys())
    assert 0.0 <= results["auc"] <= 1.0
    assert results["accuracy"] > 0.5


def test_multiclass_classification_metrics_structure():
    y_true = np.array([0, 1, 2, 1, 0, 2])
    y_prob = np.array(
        [
            [0.8, 0.1, 0.1],
            [0.2, 0.6, 0.2],
            [0.2, 0.1, 0.7],
            [0.2, 0.5, 0.3],
            [0.7, 0.2, 0.1],
            [0.1, 0.2, 0.7],
        ]
    )
    mcm = MulticlassClassificationMetrics(y_true, y_prob)
    results = mcm.compute(bootstrap=False)

    assert set(results.keys()) == {"auc (macro)", "precision (macro)", "recall (macro)"}
    assert all(0.0 <= value <= 1.0 for value in results.values())


def test_regression_metrics_basic_values():
    y_true = np.array([3.0, 4.0, 5.0, 6.0])
    y_pred = np.array([2.9, 4.1, 5.2, 5.8])
    rm = RegressionMetrics(y_true, y_pred)
    results = rm.compute(bootstrap=False)

    assert set(results.keys()) == {"mse", "mae"}
    assert results["mse"] >= 0.0
    assert results["mae"] >= 0.0



"""Metric utilities for classification and regression tasks."""

import numpy as np
from numpy.typing import ArrayLike
from sklearn.metrics import (
    roc_auc_score, average_precision_score, confusion_matrix, f1_score,
    mean_squared_error, mean_absolute_error, precision_score, recall_score
)
from .bootstrap import Bootstrapper

class BaseClassificationMetrics:
    def __init__(self, y_true: ArrayLike, y_pred_prob: ArrayLike, threshold: float = 0.5):
        self.y_true = y_true
        self.y_pred_prob = y_pred_prob
        self.threshold = threshold

    def compute(self):
        raise NotImplementedError("Must implement in subclass")

class BinaryClassificationMetrics(BaseClassificationMetrics):
    def __init__(self, y_true, y_pred_prob, threshold=0.5):
        super().__init__(y_true, y_pred_prob, threshold)
        self.y_pred = (y_pred_prob >= threshold).astype(int)

    @staticmethod
    def sensitivity(y_true,y_pred):
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
        return tp / (tp + fn)
    
    @staticmethod
    def specificity(y_true,y_pred):
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
        return tn / (tn + fp)
    
    @staticmethod
    def ppv(y_true,y_pred):
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
        return tp / (tp + fp) if (tp + fp) > 0 else 0
    
    @staticmethod
    def npv(y_true,y_pred):
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
        return tn / (tn + fn) if (tn + fn) > 0 else 0

    @staticmethod
    def accuracy(y_true, y_pred):
        tn, fp, fn, tp = confusion_matrix(y_true, y_pred).ravel()
        return (tp + tn) / (tp + tn + fp + fn)

    def compute(self, bootstrap: bool = True, n_resamples: int = 1000) -> dict:
        if bootstrap:
            bs = Bootstrapper(n_resamples=n_resamples)
            auc, auc_lower, auc_upper = bs.bootstrap_ci(roc_auc_score,y_true=self.y_true,y_pred=self.y_pred_prob)
            auprc, auprc_lower, auprc_upper =  bs.bootstrap_ci(average_precision_score,y_true=self.y_true,y_pred=self.y_pred_prob)
            f1, f1_lower, f1_upper = bs.bootstrap_ci(f1_score,y_true=self.y_true,y_pred=self.y_pred)
            sens, sens_lower, sens_upper = bs.bootstrap_ci(self.sensitivity,y_true=self.y_true,y_pred=self.y_pred)
            spec, spec_lower, spec_upper = bs.bootstrap_ci(self.specificity,y_true=self.y_true,y_pred=self.y_pred)
            ppv, ppv_lower, ppv_upper = bs.bootstrap_ci(self.ppv,y_true=self.y_true,y_pred=self.y_pred)
            npv, npv_lower,  npv_upper = bs.bootstrap_ci(self.npv,y_true=self.y_true,y_pred=self.y_pred)
            acc, acc_lower, acc_upper = bs.bootstrap_ci(self.accuracy,y_true=self.y_true,y_pred=self.y_pred)
            return {
                'auc': auc,
                'auc_upper': auc_upper,
                'auc_lower': auc_lower,
                'auprc': auprc,
                'auprc_upper': auprc_upper, 
                'auprc_lower': auprc_lower,
                'f1': f1,
                'f1_upper': f1_upper,
                'f1_lower': f1_lower,
                'sensitivity': sens,
                'sensitivity_upper': sens_upper,
                'sensitivity_lower': sens_lower,
                'specificity': spec,
                'specificity_upper': spec_upper,
                'specificity_lower': spec_lower,
                'ppv': ppv,
                'ppv_upper': ppv_upper,
                'ppv_lower': ppv_lower,
                'npv': npv,
                'npv_upper': npv_upper,
                'npv_lower': npv_lower,
                'accuracy': acc,
                'accuracy_upper': acc_upper,
                'accuracy_lower': acc_lower
            }
        else:
            tn, fp, fn, tp = confusion_matrix(self.y_true, self.y_pred).ravel()
            return {
                'auc': roc_auc_score(self.y_true, self.y_pred_prob),
                'auprc': average_precision_score(self.y_true, self.y_pred_prob),
                'f1': f1_score(self.y_true, self.y_pred),
                'sensitivity': tp / (tp + fn),
                'specificity': tn / (tn + fp),
                'ppv': tp / (tp + fp) if (tp + fp) > 0 else 0,
                'npv': tn / (tn + fn) if (tn + fn) > 0 else 0,
                'accuracy': (tp + tn) / (tp + tn + fp + fn)
            }
        
    def compute_static(self, bootstrap: bool = True, n_resamples: int = 1000) -> dict:
            if bootstrap:
                bs = Bootstrapper(n_resamples)
                f1, f1_lower, f1_upper = bs.bootstrap_ci(f1_score,y_true=self.y_true,y_pred=self.y_pred)
                sens, sens_lower, sens_upper = bs.bootstrap_ci(self.sensitivity,y_true=self.y_true,y_pred=self.y_pred)
                spec, spec_lower, spec_upper = bs.bootstrap_ci(self.specificity,y_true=self.y_true,y_pred=self.y_pred)
                ppv, ppv_lower, ppv_upper = bs.bootstrap_ci(self.ppv,y_true=self.y_true,y_pred=self.y_pred)
                npv, npv_lower,  npv_upper = bs.bootstrap_ci(self.npv,y_true=self.y_true,y_pred=self.y_pred)
                acc, acc_lower, acc_upper = bs.bootstrap_ci(self.accuracy,y_true=self.y_true,y_pred=self.y_pred)
                return {
                    'f1': f1,
                    'f1_upper': f1_upper,
                    'f1_lower': f1_lower,
                    'sensitivity': sens,
                    'sensitivity_upper': sens_upper,
                    'sensitivity_lower': sens_lower,
                    'specificity': spec,
                    'specificity_upper': spec_upper,
                    'specificity_lower': spec_lower,
                    'ppv': ppv,
                    'ppv_upper': ppv_upper,
                    'ppv_lower': ppv_lower,
                    'npv': npv,
                    'npv_upper': npv_upper,
                    'npv_lower': npv_lower,
                    'accuracy': acc,
                    'accuracy_upper': acc_upper,
                    'accuracy_lower': acc_lower
                }
            else:     
                tn, fp, fn, tp = confusion_matrix(self.y_true, self.y_pred).ravel()
                return {
                    'f1': f1_score(self.y_true,self.y_pred),
                    'sensitivity': tp / (tp + fn),
                    'specificity': tn / (tn + fp),
                    'ppv': tp / (tp + fp) if (tp + fp) > 0 else 0,
                    'npv': tn / (tn + fn) if (tn + fn) > 0 else 0,
                    'accuracy': (tp + tn) / (tp + tn + fp + fn)
                }

class MulticlassClassificationMetrics(BaseClassificationMetrics):
    def __init__(self, y_true, y_pred_prob, average: str = 'macro'):
        super().__init__(y_true, y_pred_prob)
        self.y_pred = np.argmax(y_pred_prob, axis=1)
        self.average = average

    def compute(self, bootstrap: bool = False, n_resamples: int = 1000) -> dict:
        output = {
            f'auc ({self.average})': roc_auc_score(self.y_true, self.y_pred_prob, multi_class='ovr', average=self.average),
            f'precision ({self.average})': precision_score(self.y_true, self.y_pred, average=self.average),
            f'recall ({self.average})': recall_score(self.y_true, self.y_pred, average=self.average)
        }
        if bootstrap:
            bs = Bootstrapper(n_resamples=n_resamples)
            auc, auc_lower, auc_upper = bs.bootstrap_ci(roc_auc_score,self.y_true, self.y_pred_prob,multi_class='ovr', average=self.average)
            f1, f1_lower, f1_upper = bs.bootstrap_ci(f1_score,y_true=self.y_true,y_pred=self.y_pred,average=self.average)
            precision, precision_lower, precision_upper = bs.bootstrap_ci(precision_score, self.y_true, self.y_pred, average=self.average)
            recall, recall_lower, recall_upper = bs.bootstrap_ci(recall_score, self.y_true, self.y_pred, average=self.average)
            output.update({
                f'auc ({self.average})': auc,
                f'auc ({self.average})_upper': auc_upper,
                f'auc ({self.average})_lower': auc_lower,
                f'f1 ({self.average})': f1,
                f'f1 ({self.average})_upper': f1_upper,
                f'f1 ({self.average})_lower': f1_lower,
                f'precision ({self.average})': precision,
                f'precision ({self.average})_upper': precision_upper,
                f'precision ({self.average})_lower': precision_lower,
                f'recall ({self.average})': recall,
                f'recall ({self.average})_upper': recall_upper,
                f'recall ({self.average})_lower': recall_lower
            })
        return output

class RegressionMetrics:
 
    def __init__(self, y_true, y_pred):
        self.y_true = y_true
        self.y_pred = y_pred

    def compute(self, bootstrap: bool = True, **kwargs) -> dict: 
        if bootstrap:
            bs = Bootstrapper(**kwargs)
            mse, mse_lower, mse_upper = bs.bootstrap_ci(mean_squared_error,self.y_true,self.y_pred)
            mae, mae_lower, mae_upper = bs.bootstrap_ci(mean_absolute_error,self.y_true,self.y_pred)
            return {
                'mse': mse,
                'mse_upper': mse_upper,
                'mse_lower': mse_lower,
                'mae': mae,
                'mae_upper': mae_upper,
                'mae_lower': mae_lower
            }
        else:
            return {
                'mse': mean_squared_error(self.y_true, self.y_pred),
                'mae': mean_absolute_error(self.y_true, self.y_pred),
            }



import numpy as np

from medevalkit.bootstrap import Bootstrapper


def _squared_error(yt, yp):
    return np.mean((yt - yp) ** 2)


def test_bootstrap_generates_expected_number_of_samples():
    y_true = np.array([0, 1, 1, 0, 1, 0], dtype=float)
    y_pred = np.array([0.1, 0.8, 0.7, 0.2, 0.9, 0.3], dtype=float)

    bs = Bootstrapper(n_resamples=50, random_state=7)
    scores = bs.bootstrap(_squared_error, y_true=y_true, y_pred=y_pred)

    assert len(scores) == 50
    assert all(np.isfinite(scores))


def test_bootstrap_ci_bounds_are_ordered_and_reproducible():
    y_true = np.linspace(0, 1, 20)
    noise = np.random.RandomState(0).normal(scale=0.05, size=20)
    y_pred = y_true + noise

    bs1 = Bootstrapper(n_resamples=100, random_state=42)
    bs2 = Bootstrapper(n_resamples=100, random_state=42)

    mean1, lower1, upper1 = bs1.bootstrap_ci(_squared_error, y_true=y_true, y_pred=y_pred)
    mean2, lower2, upper2 = bs2.bootstrap_ci(_squared_error, y_true=y_true, y_pred=y_pred)

    assert np.isclose(mean1, mean2)
    assert np.isclose(lower1, lower2)
    assert np.isclose(upper1, upper2)
    assert lower1 <= mean1 <= upper1



"""Tools to compare multiple models or predictions."""

import numpy as np
from numpy.typing import ArrayLike
from scipy.stats import ttest_ind, mannwhitneyu, norm
from sklearn.metrics import roc_auc_score
from .bootstrap import Bootstrapper

def _delong_roc_variance(ground_truth, predictions):
    order = np.argsort(-predictions)
    label_ordered = ground_truth[order]
    n1 = np.sum(label_ordered == 1)
    n2 = np.sum(label_ordered == 0)
    pos_scores = predictions[ground_truth == 1]
    neg_scores = predictions[ground_truth == 0]

    aucs = []
    v01 = []
    v10 = []
    for i in range(n1):
        for j in range(n2):
            if pos_scores[i] > neg_scores[j]:
                aucs.append(1)
            elif pos_scores[i] == neg_scores[j]:
                aucs.append(0.5)
            else:
                aucs.append(0)
    auc = np.mean(aucs)

    for r in range(n1):
        v01.append(np.mean([1 if pos_scores[r] > x else 0.5 if pos_scores[r] == x else 0 for x in neg_scores]))
    for r in range(n2):
        v10.append(np.mean([1 if x > neg_scores[r] else 0.5 if x == neg_scores[r] else 0 for x in pos_scores]))

    s01 = np.var(v01) / n1
    s10 = np.var(v10) / n2
    return auc, s01 + s10

def delong_test(y_true, pred1, pred2):
    auc1, var1 = _delong_roc_variance(y_true, pred1)
    auc2, var2 = _delong_roc_variance(y_true, pred2)
    se_diff = np.sqrt(var1 + var2)
    z_score = (auc1 - auc2) / se_diff
    p_value = 2 * norm.sf(abs(z_score))
    return p_value, auc1, auc2

class ModelComparer:
    def __init__(self, model_dict, x, y):
        self.model_dict = model_dict
        self.x = x
        self.y = y

    @staticmethod
    def parametric(array_a: ArrayLike, array_b: ArrayLike) -> float:
        return ttest_ind(array_a, array_b, equal_var=False).pvalue

    @staticmethod
    def nonparametric(array_a: ArrayLike, array_b: ArrayLike) -> float:
        return mannwhitneyu(array_a, array_b).pvalue

    def compare_auc(self, method: str = 'bootstrap', parametric: bool = True, n_resamples: int = None) -> dict:
    
        if parametric:
            test_func = self.parametric
        else:
            test_func = self.nonparametric

        output = {}
        model_names = list(self.model_dict.keys())

        if method == 'bootstrap':
            ##Implement pairwise t-tests using bootstrapped resamples
            aucs = {}
            bs = Bootstrapper(n_resamples=n_resamples)
            for name, model in self.model_dict.items():
                y_pred_prob = model.predict_proba(self.x)[:, 1]
                aucs[name] = bs.bootstrap(metric_fn=roc_auc_score, y_true= self.y, y_pred = y_pred_prob)

            if len(self.model_dict.keys()) < 2:
                raise ValueError('Cannot make pairwise comparisons on a single model!')
            elif len(self.model_dict.keys()) >= 2:
                for i in range(len(model_names)):
                    for j in range(i + 1, len(model_names)):
                            name1, name2 = model_names[i], model_names[j]
                            auc1 = np.median(aucs[name1])
                            auc2 = np.median(aucs[name2])
                            p = test_func(aucs[name1],aucs[name2])
                            output[f"{name1} vs {name2}"] = {
                                "AUC1": auc1,
                                "AUC2": auc2,
                                "p_value": p
                            }

        elif method == 'delong':
            for i in range(len(model_names)):
                for j in range(i + 1, len(model_names)):
                    name1, name2 = model_names[i], model_names[j]
                    pred1 = self.model_dict[name1].predict_proba(self.x)[:, 1]
                    pred2 = self.model_dict[name2].predict_proba(self.x)[:, 1]
                    p, auc1, auc2 = delong_test(self.y, pred1, pred2)
                    output[f"{name1} vs {name2}"] = {
                        "AUC1": auc1,
                        "AUC2": auc2,
                        "p_value": p
                    }
        else:
            raise ValueError(f'Input parameter method: "{method}" must be either "bootstrap", or "delong"')
        
        return output



"""High-level evaluators for model predictions."""

from __future__ import annotations

import copy
from typing import Optional, Sequence

import numpy as np
from numpy.typing import ArrayLike
import pandas as pd

from .metrics import (
    BinaryClassificationMetrics,
    MulticlassClassificationMetrics,
    RegressionMetrics,
)
from .calibration import BinaryCalibration, MulticlassCalibration
from .simulation import PopulationSimulator


def _to_numpy(array: ArrayLike) -> np.ndarray:
    """Convert arbitrary array-like input to a NumPy array."""
    if isinstance(array, (pd.Series, pd.DataFrame)):
        return array.to_numpy()
    return np.asarray(array)


class Evaluate:
    """Evaluate predictions for classification or regression tasks."""

    def __init__(
        self,
        y: ArrayLike,
        y_prob: Optional[ArrayLike] = None,
        y_pred: Optional[ArrayLike] = None,
        classification: bool = True,
        threshold: float = 0.5,
    ) -> None:
        self.y_true = _to_numpy(y)
        if self.y_true.ndim != 1:
            raise ValueError("y_true must be a one-dimensional array")

        self.threshold = threshold
        self.classification = classification
        self.results: Optional[dict] = None

        if classification:
            if y_prob is None:
                raise ValueError("y_prob must be provided for classification tasks")
            prob_array = self._prepare_probabilities(y_prob)
            if prob_array.shape[0] != self.y_true.shape[0]:
                raise ValueError("y_true and y_prob must have the same number of samples")

            self.y_pred_prob = prob_array
            self.n_classes = prob_array.shape[1]

            if self.n_classes == 2:
                positive_prob = prob_array[:, 1]
                self.y_pred = (positive_prob >= threshold).astype(int)
            else:
                self.y_pred = np.argmax(prob_array, axis=1)
        else:
            if y_pred is None:
                raise ValueError("y_pred must be provided for regression tasks")
            pred_array = _to_numpy(y_pred)
            if pred_array.shape[0] != self.y_true.shape[0]:
                raise ValueError("y_true and y_pred must have the same number of samples")

            self.y_pred = pred_array
            self.y_pred_prob = None
            self.n_classes = 0

    @staticmethod
    def _prepare_probabilities(y_prob: ArrayLike) -> np.ndarray:
        """Ensure probability inputs are formatted as a 2D NumPy array."""
        prob_array = _to_numpy(y_prob)
        if prob_array.ndim == 1:
            if np.any((prob_array < 0) | (prob_array > 1)):
                raise ValueError("Probabilities must be between 0 and 1")
            prob_array = np.column_stack([1 - prob_array, prob_array])
        elif prob_array.ndim == 2:
            if np.any((prob_array < 0) | (prob_array > 1)):
                raise ValueError("Probabilities must be between 0 and 1")
        else:
            raise ValueError("y_prob must be a 1D or 2D array-like object")
        return prob_array.astype(float)

    @staticmethod
    def try_round(value, digits: int):
        try:
            return round(value, digits)
        except Exception:
            return value

    def _normalize_indices(self, indices: Optional[Sequence[int]]) -> np.ndarray:
        if indices is None:
            return np.arange(self.y_true.shape[0], dtype=int)
        return _to_numpy(indices).astype(int)

    def construct_text_report(self, metrics: dict, bootstrap: bool, n_resamples: int) -> dict:
        metrics_with_ci = copy.deepcopy(metrics)

        for metric_category, metric_dict in metrics.items():
            metric_names_with_ci = {
                key.replace("_upper", "").replace("_lower", "")
                for key in metric_dict
                if key.endswith("_upper") or key.endswith("_lower")
            }

            for metric_name in metric_names_with_ci:
                val = np.round(metric_dict[metric_name], 3)
                upper = np.round(metric_dict[f"{metric_name}_upper"], 3)
                lower = np.round(metric_dict[f"{metric_name}_lower"], 3)
                metrics_with_ci[metric_category][metric_name] = f"{val} (95% CI {lower}-{upper})"
                metrics_with_ci[metric_category].pop(f"{metric_name}_upper", None)
                metrics_with_ci[metric_category].pop(f"{metric_name}_lower", None)

        clf_metrics = metrics_with_ci.get("clf_metrics", {"Not Applicable": np.nan})
        calib_metrics = metrics_with_ci.get("calib_metrics", {"Not Applicable": np.nan})
        reg_metrics = metrics_with_ci.get("reg_metrics", {"Not Applicable": np.nan})

        report = (
            "==== Report ====\n"
            f"Bootstrap: {bootstrap}\n"
            f"Resamples: {n_resamples}\n\n"
            "-- Classification Metrics --\n"
            + "\n".join(
                f"{key}: {self.try_round(value, 3)}" for key, value in clf_metrics.items()
            )
            + "\n\n-- Calibration Metrics --\n"
            + "\n".join(
                f"{key}: {self.try_round(value, 3)}"
                for key, value in calib_metrics.items()
                if "calibration_curve" not in key
            )
            + "\n\n-- Regression Metrics --\n"
            + "\n".join(
                f"{key}: {self.try_round(value, 3)}" for key, value in reg_metrics.items()
            )
        )

        return {"text_report": report}

    def generate_report(
        self,
        indices: Optional[Sequence[int]] = None,
        multiclass_method: str = "ovr",
        calibration_bins: int = 10,
        bootstrap: bool = True,
        n_resamples: int = 1000,
        **kwargs,
    ) -> dict:
        indices_array = self._normalize_indices(indices)

        clf_metrics = {}
        calib_metrics = {}
        reg_metrics = {}

        if self.classification:
            if self.n_classes > 2:
                subset_prob = self.y_pred_prob[indices_array, :]
                subset_true = self.y_true[indices_array]
                clf_metrics = MulticlassClassificationMetrics(subset_true, subset_prob).compute(
                    bootstrap=bootstrap,
                    n_resamples=n_resamples,
                    **kwargs,
                )
                mcc = MulticlassCalibration(subset_true, subset_prob, n_bins=calibration_bins)
                if multiclass_method == "ovr":
                    ovr_calib = mcc.one_vs_rest_curves(
                        bootstrap=bootstrap,
                        n_resamples=n_resamples,
                        **kwargs,
                    )
                    for class_label, class_metrics in ovr_calib.items():
                        for key, value in class_metrics.items():
                            calib_metrics[f"{class_label}_{key}"] = value
                elif multiclass_method == "ece":
                    calib_metrics = mcc.expected_calibration_error(
                        bootstrap=bootstrap,
                        n_resamples=n_resamples,
                        **kwargs,
                    )
                else:
                    raise ValueError(
                        "Parameter `multiclass_method` must be one of ['ovr','ece']"
                    )
            else:
                positive_prob = self.y_pred_prob[indices_array, 1]
                subset_true = self.y_true[indices_array]
                clf_metrics = BinaryClassificationMetrics(
                    subset_true,
                    positive_prob,
                    threshold=self.threshold,
                ).compute(bootstrap=bootstrap, n_resamples=n_resamples, **kwargs)
                calib_metrics = BinaryCalibration(
                    subset_true,
                    positive_prob,
                    n_bins=calibration_bins,
                ).compute(bootstrap=bootstrap, n_resamples=n_resamples, **kwargs)

            output = {"clf_metrics": clf_metrics, "calib_metrics": calib_metrics}
        else:
            subset_true = self.y_true[indices_array]
            subset_pred = self.y_pred[indices_array]
            reg_metrics = RegressionMetrics(subset_true, subset_pred).compute(
                bootstrap=bootstrap,
                n_resamples=n_resamples,
                **kwargs,
            )
            output = {"reg_metrics": reg_metrics}

        report = self.construct_text_report(output, bootstrap, n_resamples)
        output.update(report)
        self.results = output
        return output


class EvaluateWithSimulation(Evaluate):
    """Evaluate predictions under varying outcome incidence via simulation."""

    def __init__(self, y_true: ArrayLike, y_prob: ArrayLike, threshold: float = 0.5) -> None:
        super().__init__(
            y_true=y_true,
            y_prob=y_prob,
            classification=True,
            threshold=threshold,
        )
        if self.n_classes != 2:
            raise ValueError("EvaluateWithSimulation currently supports binary classification only")
        self.results = None

    def run_simulation(
        self,
        incidence_rate_list: Optional[Sequence[float]] = None,
        calibration_bins: int = 10,
        bootstrap: bool = True,
        n_resamples: int = 1000,
        random_state: int = 1,
    ) -> dict:
        if incidence_rate_list is None:
            incidence_rate_list = [
                0.01,
                0.05,
                0.1,
                0.2,
                0.3,
                0.4,
                0.5,
                0.6,
                0.7,
                0.8,
                0.9,
                0.95,
                0.99,
            ]

        simulator = PopulationSimulator(self.y_true, n_resamples, random_state)
        self.sim_dict = simulator.simulate(list(incidence_rate_list))
        self.sim_results = {}

        for incidence, indices in self.sim_dict.items():
            metrics = self.generate_report(
                indices=indices,
                calibration_bins=calibration_bins,
                bootstrap=bootstrap,
                n_resamples=n_resamples,
            )

            combined_metrics = {}
            for key, value in metrics.items():
                if isinstance(value, dict):
                    combined_metrics.update(value)
                else:
                    combined_metrics[key] = value

            summary = {}
            for metric_name, metric_value in combined_metrics.items():
                if any(token in metric_name for token in ("upper", "lower", "text", "curve")):
                    continue
                values = [metric_value]
                mean_val = np.mean(values)
                ci_low, ci_high = np.percentile(values, [2.5, 97.5])
                summary[metric_name] = f"{mean_val:.3f} (95% CI {ci_low:.3f}-{ci_high:.3f})"

            self.sim_results[incidence] = summary

        return self.sim_results

    def generate_metrics(self) -> pd.DataFrame:
        output = pd.DataFrame.from_dict(self.sim_results, orient="index").T
        output.columns = [np.round(col, 2) for col in output.columns]
        return output
